"""Localization utilities to translate error messages"""

import sys

translations = {
    "INVALID_SIZE": {
        "en": "Invalid size (must be a positive number)",
        "it": "Dimensione non valida (deve essere un numero positivo)",
    },
    "EMPTY_GRAPHICS_LIST": {
        "en": "The list of graphics cannot be empty",
        "it": "La lista delle grafiche non può essere vuota"
    }
}


def translate(key: str) -> str:
    """Looks up an error message by `key` and returns the localized version.

    :param key: error message key
    :returns: locale-specific descriptive error message
    """
    lang = sys.modules["pytamaro"].LANGUAGE
    return translations[key][lang]



"""
Der Typ `Farbe`, Funktionen, die Farben erzeugen
und Konstanten für wichtige Farben.
"""

from __future__ import annotations

from pytamaro.color import Color, rgb_color, rgba_color

Farbe = Color
"""
Repräsentiert eine Farbe.
Eine Farbe hat auch eine gewisse Durchsichtigkeit,
von komplett undurchsichtig, wie die Farbe `rot`,
bis komplett durchsichtig, wie die Farbe `transparent`.
"""


def rgb_farbe(rot: int, gruen: int, blau: int) -> Farbe:
    """
    Erzeugt eine komplett opake (undurchsichtige) Farbe
    mit den gegebenen Rot-, Grün- und Blau-Komponenten (RGB).

    :param rot: die rote Farb-Komponente (0 -- 255)
    :param gruen: die grüne Farb-Komponente (0 -- 255)
    :param blau: die blaue Farb-Komponente (0 -- 255)
    :returns: eine Farbe mit den gegebenen RGB-Komponenten
    """
    return rgb_color(rot, gruen, blau)


def rgba_farbe(rot: int, gruen: int, blau: int, alpha: int) -> Farbe:
    """
    Erzeugt eine Farbe mit den gegebenen Komponenten für Rot, Grün, Blau,
    und die Opazität (Undurchsichtigkeit) `alpha` (RGBA).

    :param rot: die rote Farb-Komponente (0 -- 255)
    :param gruen: die grüne Farb-Komponente (0 -- 255)
    :param blau: die blaue Farb-Komponente (0 -- 255)
    :param alpha: die alpha-Komponente (Undurchsichtigkeit, Opazität),
                  wobei 0 komplett durchsichtig und 255 komplett undurchsichtig
                  entspricht
    :returns: eine Farbe mit den gegebenen RGBA-Komponenten
    """
    return rgba_color(rot, gruen, blau, alpha)



"""
Functions to do I/O with graphics, such as showing or saving them.
"""

from typing import List

from skia import kPNG

from pytamaro.debug import add_debug_info
from pytamaro.graphic import Graphic
from pytamaro.graphic_utils import graphic_to_image
from pytamaro.localization import translate
from pytamaro.utils import export, is_notebook


@export
def show_graphic(graphic: Graphic, debug: bool = False):
    """
    Show a graphic in a window. Empty graphics cannot be shown and produce no
    effect when shown using this function.

    When `debug` is `True`, adorns the visualization with useful information
    for debugging: a red border around the bounding box and a yellowish cross
    around the pinning position.

    :param graphic: graphic to be shown
    :param debug: can be optionally set to `True` to overlay debugging
           information
    """
    if not graphic.is_empty_graphic():
        to_show = add_debug_info(graphic) if debug else graphic
        pil_image = graphic_to_image(to_show)
        if is_notebook():
            # pylint: disable=undefined-variable
            display(pil_image)  # type: ignore[name-defined]
        else:
            pil_image.show()


@export
def save_graphic(filename: str, graphic: Graphic, debug: bool = False):
    """
    Save a graphic as a PNG file. Empty graphics cannot be saved and produce no
    effect when saved using this function.

    When `debug` is `True`, adorns the visualization with useful information
    for debugging: a red border around the bounding box and a yellowish cross
    around the pinning position.

    :param filename: name of the file to create (without the extension)
    :param graphic: graphic to be saved
    :param debug: can be optionally set to `True` to overlay debugging
           information
    """
    if not graphic.is_empty_graphic():
        to_show = add_debug_info(graphic) if debug else graphic
        image = to_show.as_image()
        image.save(f"{filename}.png", kPNG)


@export
def save_gif(
    filename: str, graphics: List[Graphic], duration: int = 40, loop: bool = True
):
    """
    Save a sequence of graphics as an aminated GIF.

    Graphics are sequentially reproduced (normally at 25 frames per second) in
    a loop (unless specificied otherwise).

    :param filename: name of the file to create (without the extension)
    :param graphics: list of graphics to be saved as a GIF
    :param duration: duration in milliseconds for each frame
           (defaults to 40 milliseconds, which leads to 25 frames per second)
    :param loop: whether the GIF should loop indefinitely (defaults to true)
    """
    if len(graphics) == 0:
        raise ValueError(translate("EMPTY_GRAPHICS_LIST"))
    pil_images = list(map(graphic_to_image, graphics))
    pil_images[0].save(
        f"{filename}.gif",
        save_all=True,
        append_images=pil_images[1:],
        duration=duration,
        loop=0 if loop else 1,  # loop 0 means "indefinitely", 1 means "once"
    )



"""
Funzioni per operazioni con grafiche (principlamente per combinarle).
"""

from __future__ import annotations

from pytamaro.it.graphic import Grafica
from pytamaro.operations import (above, beside, compose, graphic_height,
                                 graphic_width, overlay, pin, rotate)


def larghezza_grafica(grafica: Grafica) -> int:
    """
    Ritorna la larghezza di una grafica, in pixel.

    :param grafica: grafica di cui calcolare la larghezza
    :returns: larghezza della grafica
    """
    return graphic_width(grafica)


def altezza_grafica(grafica: Grafica) -> int:
    """
    Ritorna l'altezza di una grafica, in pixel.

    :param grafica: grafica di cui calcolare l'altezza
    :returns: altezza della grafica
    """
    return graphic_height(grafica)


def componi(grafica_primopiano: Grafica, grafica_secondopiano: Grafica) \
        -> Grafica:
    """
    Compone due grafiche tenendo la prima in primo piano e la seconda sullo
    sfondo, allineandole usando le loro posizioni di fissaggio.

    La posizione di fissaggio usata per comporre diventa la posizione di
    fissaggio della grafica risultante.

    :param grafica_primopiano: grafica da tenere in primo piano
    :param grafica_secondopiano: grafica da tenere sullo sfondo
    :returns: grafica risultante dalla composizione delle due fornite
    """
    return compose(grafica_primopiano, grafica_secondopiano)


def fissa(posizione_orizzontale: str, posizione_verticale: str,
          grafica: Grafica) -> Grafica:
    """
    Cambia la posizione di fissaggio di una grafica, ritornando una nuova
    grafica con lo stesso contenuto ma una posizione di fissaggio aggiornata.

    La nuova posizione di fissaggio è determinata dai parametri
    `posizione_orizzontale` e `posizione_orizzontale`.

    :param posizione_orizzontale: uno tra "sinistra", "centro", "destra" per
           muovere la nuova posizione di fissaggio rispettivamente al bordo
           sinistro, al centro (orizzontalmente), o al bordo destro della
           grafica
    :param posizione_verticale: uno tra "alto", "centro", "basso" per
           muovere la nuova posizione di fissaggio rispettivamente al bordo
           superiore, al centro (verticalmente), o al bordo inferiore della
           grafica
    :param grafica: grafica originale
    :returns: una nuova grafica con una posizione di fissaggio aggiornata
    """
    h_mapping = {
        "sinistra": "left",
        "centro": "middle",
        "destra": "right"
    }
    v_mapping = {
        "alto": "top",
        "centro": "middle",
        "basso": "bottom"
    }
    return pin(h_mapping[posizione_orizzontale],
               v_mapping[posizione_verticale],
               grafica)


def sovrapponi(grafica_primopiano: Grafica, grafica_secondopiano: Grafica) \
        -> Grafica:
    """
    Sovrappone due grafiche tenendo la prima in primo piano e la seconda sullo
    sfondo, allineandole sui loro centri.

    :param grafica_primopiano: grafica da tenere in primo piano
    :param grafica_secondopiano: grafica da tenere sullo sfondo
    :returns: grafica risultante dalla sovrapposizione delle due fornite
    """
    return overlay(grafica_primopiano, grafica_secondopiano)


def accanto(grafica_sinistra: Grafica, grafica_destra: Grafica) -> Grafica:
    """
    Compone due grafiche affiancandole, posizionando la prima a sinistra e la
    seconda a destra. Le due grafiche vengono allineate verticalmente al
    centro.

    :param grafica_sinistra: grafica da posizionare a sinistra
    :param grafica_destra: grafica da posizionare a destra
    :returns: grafica risultante dall'affiancamento orizzontale delle due
              grafiche fornite
    """
    return beside(grafica_sinistra, grafica_destra)


def sopra(grafica_alto: Grafica, grafica_basso: Grafica) -> Grafica:
    """
    Compone due grafiche affiancandole verticalmente, posizionando la prima in
    alto e la seconda in basso. Le due grafiche vengono allineate
    orizzontalmente al centro.

    :param grafica_sopra: grafica da posizionare sopra
    :param grafica_sotto: grafica da posizionare sotto
    :returns: grafica risultante dall'affiancamento verticale delle due
              grafiche fornite
    """
    return above(grafica_alto, grafica_basso)


def ruota(gradi: float, grafica: Grafica) -> Grafica:
    """
    Ruota una grafica di un certo numero di gradi in senso antiorario attorno
    alla sua posizione di fissaggio.

    È possibile che si verifichino piccoli errori di arrotondamento (a causa
    dell'approssimazione al pixel più vicino).

    :param gradi: numero di gradi di cui ruotare la grafica
    :param grafica: grafica da ruotare
    :returns: la grafica originale ruotata attorno alla sua posizione di
              fissaggio
    """
    return rotate(gradi, grafica)



"""
Graphic class, which wraps a Pillow image with a position for pinning.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import accumulate
from math import sqrt
from typing import List

from skia import (Canvas, Font, Image, Matrix, Paint, Path, Point, Rect, Size,
                  Surface, Typeface)

from pytamaro.color import Color
from pytamaro.color_names import black

# pylint: disable=super-init-not-called


@dataclass
class Graphic(ABC):
    """
    A graphic (image) with a position for pinning.

    The pinning position is used in the following operations:

    - rotation (to determine the center of rotation)
    - graphic composition (two graphics get composed aligning their pinning
      position).
    """
    pin_position: Point
    path: Path

    def set_pin_position(self, x: float, y: float):  # pylint: disable=invalid-name
        """
        Changes the pinning position of this graphic.

        :param x: x coordinate
        :param y: y coordinate
        """
        self.pin_position = Point(x, y)

    def size(self) -> Size:
        """
        Computes the size of this graphic (x and y axes spanning),
        using the bounds computed by bounds().

        :returns: graphic's size
        """
        bounds = self.bounds()
        return Size(bounds.width(), bounds.height())

    def bounds(self) -> Rect:
        """
        Computes the (tight) bounds for the path (outline) of this graphic.

        :returns: a rectangle that indicates the bounds of the graphic in the 2D
                  space
        """
        return self.path.computeTightBounds()

    @abstractmethod
    def render(self, canvas: Canvas):
        """
        Renders the current graphic onto the provided canvas.

        :param canvas: canvas onto which to render
        """

    def is_empty_graphic(self) -> bool:  # pylint: disable=no-self-use
        """
        Returns whether this graphic is empty or not.

        :returns: True if the graphic is empty, False otherwise
        """
        return False

    def as_image(self) -> Image:
        """
        Renders this graphic as an image.

        The image has the minimal size required to fit the graphic.

        :returns: an image with the rendered graphic
        """
        int_size = self.size().toCeil()
        surface = Surface(int_size.width(), int_size.height())
        with surface as canvas:
            bounds = self.bounds()
            canvas.translate(-bounds.left(), -bounds.top())
            self.render(canvas)
        return surface.makeImageSnapshot()

    def _key(self) -> List:
        """
        Returns the fields relevant for __eq__ and __hash__ as a list.
        The resulting list must contain comparable and hashable values.
        For this reason, we carefully manage the case with the empty image.

        :returns: a list with values relevant for equality
        """
        if self.is_empty_graphic():
            return [None]

        bitmap = self.as_image().bitmap()
        pixels = [bitmap.getColor(x, y)
                  for y in range(bitmap.height()) for x in range(bitmap.width())]
        return [tuple(pixels)]

    def __eq__(self, other: object) -> bool:
        """
        Compares a graphic with another graphic.
        Two graphics are considered equal if they are both empty or if they
        have exactly the same content when rendered (pixel by pixel)/

        :returns: True if the two graphics are considered equal, False
                  otherwise
        """
        if not isinstance(other, Graphic):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self) -> int:
        """
        Computes the hash of this graphic, relying on _key() and the built-in
        hash().

        :returns hash for this graphic
        """
        return hash(tuple(self._key()))


class Primitive(Graphic):
    """
    Represents a primitive graphic, which has a path and a uniform color.
    Geometric shapes and text are primitive graphics.
    """
    def __init__(self, path: Path, color: Color):
        self.path = path
        self.paint = Paint(color.color)
        bounds = self.path.computeTightBounds()
        self.set_pin_position(bounds.width() / 2, bounds.height() / 2)

    def render(self, canvas: Canvas):
        canvas.drawPath(self.path, self.paint)


class Empty(Primitive):
    """
    An empty graphic.
    """
    def __init__(self):
        super().__init__(Path(), black)

    def is_empty_graphic(self) -> bool:
        return True


class Rectangle(Primitive):
    """
    A rectangle.
    """
    def __init__(self, width: float, height: float, color: Color):
        path = Path().addRect(Rect.MakeWH(width, height))
        super().__init__(path, color)


class Ellipse(Primitive):
    """
    An ellipse.
    """
    def __init__(self, width: float, height: float, color: Color):
        path = Path().addOval(Rect.MakeWH(width, height))
        super().__init__(path, color)


class CircularSector(Primitive):
    """
    A circular sector (with an angle between 1 and 359).
    """
    def __init__(self, radius: float, angle: float, color: Color):
        circle = Rect.MakeWH(2 * radius, 2 * radius)
        path = Path()
        path.moveTo(radius, radius)
        path.arcTo(circle, 0, angle, False)
        path.close()
        super().__init__(path, color)


class Triangle(Primitive):
    """
    An upwards-pointing equilateral triangle.
    """
    def __init__(self, side: float, color: Color):
        height = side * sqrt(3) / 2
        points = [Point(0, 0), Point(-side/2, height), Point(side/2, height)]
        path = Path().addPoly(points, close=True)
        super().__init__(path, color)


class Text(Primitive):
    """
    Graphic containing text, using a given font with a given typographic size.
    """
    def __init__(self, text: str, font_name: str, size: float, color: Color):
        font = Font(Typeface(font_name), size)
        glyphs = font.textToGlyphs(text)
        offsets = [0] + list(accumulate(font.getWidths(glyphs)))
        text_path = Path()
        for glyph, x_offset in zip(glyphs, offsets):
            path = font.getPath(glyph)
            # Some glyphs (e.g., the space) have no outline
            if path is not None:
                path.offset(x_offset, 0)
                text_path.addPath(path)
        super().__init__(text_path, color)


class Compose(Graphic):
    """
    Represents the composition of two graphics, one in the foreground and the
    other in the background, joined on their pinning positions.
    """
    def __init__(self, foreground: Graphic, background: Graphic):
        self.foreground = foreground
        self.background = background
        fg_pin = self.foreground.pin_position
        bg_pin = self.background.pin_position
        self.set_pin_position(bg_pin.x(), bg_pin.y())
        self.path = Path(self.background.path)
        self.path.addPath(self.foreground.path,
                          bg_pin.x() - fg_pin.x(), bg_pin.y() - fg_pin.y())

    def render(self, canvas: Canvas):
        canvas.save()
        self.background.render(canvas)
        canvas.translate(self.background.pin_position.x() - self.foreground.pin_position.x(),
                         self.background.pin_position.y() - self.foreground.pin_position.y())
        self.foreground.render(canvas)
        canvas.restore()


class Pin(Graphic):
    """
    Represents the pinning of a graphic in a certain position on its bounds.
    """
    def __init__(self, graphic: Graphic, horizontal_place: str, vertical_place: str):
        self.graphic = graphic
        bounds = self.graphic.bounds()
        h_mapping = {
            "left": bounds.left(),
            "middle": bounds.centerX(),
            "right": bounds.right()
        }
        v_mapping = {
            "top": bounds.top(),
            "middle": bounds.centerY(),
            "bottom": bounds.bottom()
        }
        self.set_pin_position(
            h_mapping[horizontal_place], v_mapping[vertical_place])
        self.path = Path(self.graphic.path)

    def render(self, canvas: Canvas):
        self.graphic.render(canvas)


class Rotate(Graphic):
    """
    Represents the rotation of a graphic by a certain angle clockwise.
    """
    def __init__(self, graphic: Graphic, deg: float):
        self.graphic = graphic
        self.rot_matrix = Matrix.RotateDeg(deg, self.graphic.pin_position)
        self.path = Path()
        self.graphic.path.transform(self.rot_matrix, self.path)  # updates self.path
        self.pin_position = graphic.pin_position

    def render(self, canvas: Canvas):
        canvas.save()
        canvas.concat(self.rot_matrix)
        self.graphic.render(canvas)
        canvas.restore()



"""
Funktionen zum Erzeugen primitiver Grafiken (Figuren und Texte)
"""

from __future__ import annotations

from pytamaro.de.color import Farbe
from pytamaro.de.graphic import Grafik
from pytamaro.primitives import (circular_sector, empty_graphic,
                                 rectangle, triangle)
from pytamaro.primitives import (ellipse as ellipse_e, text as text_e)


def rechteck(breite: float, hoehe: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt ein Rechteck mit der gegebenen Breite und Höhe,
    gefüllt in der gegebenen Farbe.

    :param breite: die Breite des Rechtecks, in Pixel
    :param hoehe: die Höhe des Rechtecks, in Pixel
    :param farbe: Füllfarbe des Rechtecks
    :returns: eine Grafik mit dem gegebenen Rechteck
    """
    return rectangle(breite, hoehe, farbe)


def leere_grafik() -> Grafik:
    """
    Erzeugt eine leere Grafik.
    Wenn eine leere Grafik mit einer anderen Grafik kombiniert wird
    verhält sie sich als neutrales Element:
    das Ergebnis der Komposition ist einfach gleich der anderen Grafik.

    Eine leere Grafik kann weder angezeigt noch gespeichert werden.

    :returns: eine leere Grafik (Breite und Höhe sind 0 Pixel)
    """
    return empty_graphic()


def ellipse(breite: float, hoehe: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt eine Ellipse mit der gegebenen Breite und Höhe,
    gefüllt in der gegebenen Farbe.

    Wenn Breite und Höhe gleich gross sind wird die Ellipse zum Kreis
    mit dem entsprechenden Durchmesser.

    :param breite: Breite der Ellipse, in Pixel
    :param hoehe: Höhe der Ellipse, in Pixel
    :param farbe: Füllfarbe der Ellipse
    :returns: eine Grafik mit dem gegebenen Rechteck
    """
    return ellipse_e(breite, hoehe, farbe)


def kreis_sektor(radius: float, winkel: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt einen Kreissektor mit dem gegebenen Radius,
    der den gegebenen Winkel umspannt,
    gefüllt in der gegebenen Farbe.

    Ein Kreissektor ist ein Teil eines Kreises
    begrenzt durch zwei Radien und einen Bogen.
    Wenn man den Kreis als Uhr betrachtet
    dann zeigt der erste Radius in Richtung 3 Uhr.
    Der Winkel bestimmt die Position des zweiten Radius,
    ausgehend vom ersten Radius im Uhrzeigersinn.

    :param radius: Kreisradius, in Pixel
    :param winkel: Winkel des Sektors, in Grad
    :param farbe: Füllfarbe des Kreissektors
    :returns: eine Grafik mit dem gegebenen Kreissektor
    """
    return circular_sector(radius, winkel, farbe)


def dreieck(seite: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt ein gleichseitiges Dreieck mit der gegebenen Seitenlänge
    und einer nach oben zeigenden Ecke,
    gefüllt in der gegebenen Farbe.

    :param seite: Seitenlänge des Dreiecks, in Pixel
    :param farbe: Füllfarbe des Dreiecks
    :returns: eine Grafik mit dem gegebenen Dreieck
    """
    return triangle(seite, farbe)


def text(inhalt: str, schriftart: str, punkte: float, farbe: Farbe) -> Grafik:
    """
    Erzeugt einen Text in der gegebenen Schriftart und Schriftgrösse,
    gefüllt in der gegebenen Farbe.

    Falls für die gegebene Schriftart auf dem System keine True-Type Schrift
    zur Verfügung steht, wird eine einfache Standardschriftart verwendet.
    Die resultierende Grafik hat die minimale Grösse,
    die den gesamten Text umschliesst.

    :param inhalt: der Text, der dargestellt werden soll
    :param schriftart: der Name der Schriftart
                       (zum Beispiel "arial" auf Windows, "Arial" auf macOS)
    :param punkte: Schriftgrösse in typografischen Punkten (zum Beispiel 16)
    :param farbe: Farbe, in der der Text dargestellt werden soll
    :returns: eine Grafik bestehend aus dem gegebenen Text
    """
    return text_e(inhalt, schriftart, punkte, farbe)



"""
Functions to create primitive graphics (shapes and text)
"""

from pytamaro.color import Color
from pytamaro.graphic import (CircularSector, Ellipse, Empty, Graphic,
                              Rectangle, Text, Triangle)
from pytamaro.graphic_utils import ensure_size
from pytamaro.utils import export


@export
def rectangle(width: float, height: float, color: Color) -> Graphic:
    """
    Creates a rectangle of the given size, filled with a color.

    :param width: width of the rectangle, in pixel
    :param height: height of the rectangle, in pixel
    :param color: the color to be used to fill the rectangle
    :returns: the specified rectangle as a graphic
    """
    ensure_size(width)
    ensure_size(height)
    return Rectangle(width, height, color)


@export
def empty_graphic() -> Graphic:
    """
    Creates an empty graphic.
    When an empty graphic is composed with any other graphic, it behaves
    as a neutral element: the result is always identical to the other graphic.

    An empty graphic cannot be shown nor saved.

    :returns: an empty graphic (width and height 0 pixels)
    """
    return Empty()


@export
def ellipse(width: float, height: float, color: Color) -> Graphic:
    """
    Creates an ellipse with the given width and height, filled with a color.

    When width and height are the same, the ellipse becomes a circle with a
    diameter equal to the provided size.

    :param width: width of the ellipse, in pixel
    :param height: height of the ellipse, in pixel
    :param color: the color to be used to fill the circle
    :returns: the specified circle as a graphic
    """
    ensure_size(width)
    ensure_size(height)
    return Ellipse(width, height, color)


@export
def circular_sector(radius: float, angle: float, color: Color) \
        -> Graphic:
    """
    Creates a circular sector belonging to a circle of the given radius, filled
    with a color.

    A circular sector is a portion of a circle enclosed between two radii and
    an arc.
    Considering a circle as a clock, the first radius is supposed to "point"
    towards 3 o'clock. The `angle` determines the position of the second
    radius, computed starting from the first one in the clockwise direction.

    :param radius: radius of the circle from which the circular sector is
                   taken, in pixel
    :param angle: central angle, in degrees
    :param color: the color to be used to fill the circular sector
    :returns: the specified circular sector as a graphic
    """
    ensure_size(radius)
    return CircularSector(radius, angle, color)


@export
def triangle(side: float, color: Color) -> Graphic:
    """
    Creates an equilateral triangle pointing upwards of the given side, filled
    with a color.

    :param side: length of the side of the triangle, in pixel
    :param color: the color to be used to fill the triangle
    :returns: the specified triangle as a graphic
    """
    ensure_size(side)
    return Triangle(side, color)


@export
def text(content: str, font: str, points: float, color: Color) \
        -> Graphic:
    """
    Creates a graphic with the text rendered using the specified font, size and
    color.

    When the indicated True-Type Font is not found in the system, a very
    basilar font that is always available is used instead. The resulting
    graphic has the minimal size that still fits the whole text.

    :param content: the text to render
    :param font: the name of the font (e.g., "arial" on Windows, "Arial" on
           macOS)
    :param points: size in typographic points (e.g., 16)
    :param color: the color to be used to render the text
    :returns: the specified text as a graphic
    """
    return Text(content, font, points, color)



"""
Functions to do operations on graphics (mainly, to combine them).
"""


from pytamaro.graphic import Compose, Graphic, Pin, Rotate
from pytamaro.utils import export


@export
def graphic_width(graphic: Graphic) -> int:
    """
    Returns the width of a graphic, in pixel.

    :param graphic: graphic to calculate the width of
    :returns: width of the graphic
    """
    return graphic.size().toCeil().width()


@export
def graphic_height(graphic: Graphic) -> int:
    """
    Returns the height of a graphic, in pixel.

    :param graphic: graphic to calculate the height of
    :returns: height of the graphic
    """
    return graphic.size().toCeil().height()


@export
def compose(foreground_graphic: Graphic, background_graphic: Graphic) \
        -> Graphic:
    """
    Composes two graphics keeping the first one in the foreground and the
    second one in background, aligning them using their pin positions.

    The pin location used to compose becomes the pin location of the resulting
    graphic.

    :param foreground_graphic: graphic to keep in the foreground
    :param background_graphic: graphic to keep in the background
    :returns: the resulting graphic after combining the two provided ones
    """
    return Compose(foreground_graphic, background_graphic)


@export
def pin(horizontal_place: str, vertical_place: str, graphic: Graphic) \
        -> Graphic:
    """
    Changes the pinning position of a graphic, returning a new graphic with
    the same content but with an updated pinning position.

    The new pinning position is determined by the parameters `horizontal_place`
    and `vertical_place`.

    :param horizontal_place: one of "left", "middle" or "right" that
           respectively indicate to move the new pinning positon to the left
           border, the (horizontal) center, or the right border of the graphic
    :param vertical_place: one of "top", "middle" or "bottom" that
           respectively indicate to move the new pinning positon to the top
           border, the (vertical) center, or the bottom border of the graphic
    :param graphic: original graphic
    :returns: a new graphic with an updated pinning position
    """
    return Pin(graphic, horizontal_place, vertical_place)


@export
def overlay(foreground_graphic: Graphic, background_graphic: Graphic) \
        -> Graphic:
    """
    Overlays two graphics keeping the first one in the foreground and the
    second one in background, aligning them on their centers.

    :param foreground_graphic: graphic to keep in the foreground
    :param background_graphic: graphic to keep in the background
    :returns: the resulting graphic after overlaying the two provided ones
    """
    return compose(
        pin("middle", "middle", foreground_graphic),
        pin("middle", "middle", background_graphic))


@export
def beside(left_graphic: Graphic, right_graphic: Graphic) -> Graphic:
    """
    Composes two graphics placing the first on the left and the second on the
    right.  The two graphics are aligned vertically on their centers.

    :param left_graphic: graphic to place on the left
    :param right_graphic: graphic to place on the right
    :returns: the resulting graphic after placing the two graphics one besides
              the other
    """
    return compose(
        pin("right", "middle", left_graphic),
        pin("left", "middle", right_graphic))


@export
def above(top_graphic: Graphic, bottom_graphic: Graphic) -> Graphic:
    """
    Composes two graphics placing the first on the top and the second on the
    bottom.
    The two graphics are aligned horizontally on their centers.

    :param top_graphic: graphic to place on the top
    :param bottom_graphic: graphic to place on the bottom
    :returns: the resulting graphic after placing the two graphics one above
              the other
    """
    return compose(
        pin("middle", "bottom", top_graphic),
        pin("middle", "top", bottom_graphic))


@export
def rotate(degrees: float, graphic: Graphic) -> Graphic:
    """
    Rotates an graphic by a given amount of degrees counterclockwise around
    its pinning position.

    Small rounding errors (due to approximations to the nearest pixel) may
    occur.

    :param degrees: amount of degrees the graphic needs to be rotated
    :param graphic: the graphic to rotate
    :returns: the original graphic rotated around its pinning position
    """
    # Negate the angle given that Rotate is clockwise.
    return Rotate(graphic, -degrees)



"""
`Color` type and functions to produce colors.
"""

from dataclasses import dataclass
from typing import Tuple

from skia import Color4f


@dataclass
class Color:
    """
    Represents a color in the RGBA color space (using integers between 0 and
    255).
    """
    color: Color4f

    def __init__(self, red: int, green: int, blue: int, alpha: int):
        self.color = Color4f(red / 255, green / 255, blue / 255, alpha / 255)

    def as_tuple(self) -> Tuple[int, int, int, int]:
        """
        Returns the current color as an RGBA tuple.

        :meta private:
        :returns: a tuple with four components (0 -- 255) that identifies this
                  color
        """
        return tuple(val * 255 for val in tuple(self.color))


def rgb_color(red: int, green: int, blue: int) -> Color:
    """
    Returns a fully-opaque color with the provided components for red, green
    and blue.

    :param red: red component (0 -- 255)
    :param green: green component (0 -- 255)
    :param blue: blue component (0 -- 255)
    :returns: a color with the provided RGB components
    """
    return rgba_color(red, green, blue, 255)


def rgba_color(red: int, green: int, blue: int, alpha: int) -> Color:
    """
    Returns a color with the provided components for red, green and blue and a
    certain degree of transparency controlled by `alpha`.

    :param red: red component (0 -- 255)
    :param green: green component (0 -- 255)
    :param blue: blue component (0 -- 255)
    :param alpha: alpha (transparency) component where 0 means fully
           transparent and 255 fully opaque
    :returns: a color with the provided RGBA components
    """
    return Color(red, green, blue, alpha)



"""
Funktionen für Operationen mit Grafiken (hauptsächlich für deren Komposition).
"""

from __future__ import annotations

from pytamaro.de.graphic import Grafik
from pytamaro.operations import (above, beside, compose, graphic_height,
                                 graphic_width, overlay, pin, rotate)


def grafik_breite(grafik: Grafik) -> int:
    """
    Gibt die Breite (in Pixel) der gegebenen Grafik zurück.

    :param grafik: Grafik deren Breite gesucht ist
    :returns: Breite der Grafik
    """
    return graphic_width(grafik)


def grafik_hoehe(grafik: Grafik) -> int:
    """
    Gibt die Höhe (in Pixel) der gegebenen Grafik zurück.

    :param grafik: Grafik deren Höhe gesucht ist
    :returns: Höhe der Grafik
    """
    return graphic_height(grafik)


def kombiniere(vordere_grafik: Grafik, hintere_grafik: Grafik) \
        -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus der Kombination der zwei gegebenen Grafiken besteht.
    Die erste gegebene Grafik liegt im Vordergrund
    und die zweite im Hintergrund.
    Die Grafiken werden so ausgerichtet,
    dass ihre Fixierungspositionen übereinanderliegen.

    Die überlappenden Fixierungspositionen werden zur Fixierungsposition der
    resultierenden Grafik.

    :param vordere_grafik: Grafik im Vordergrund
    :param hintere_grafik: Grafik im Hintergrund
    :returns: die zusammengesetzte Grafik
    """
    return compose(vordere_grafik, hintere_grafik)


def fixiere(horizontale_position: str, vertikale_position: str,
            grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik, die der gegebenen Grafik
    mit einer anderen Fixierungsposition entspricht.

    Die neue Fixierungsposition wird mit den Parametern
    `horizontale_position` und `vertikale_position` bestimmt.

    :param horizontale_position: "links", "mitte" oder "rechts" um
           die Fixierungsposition auf den linken Rand, in die Mitte,
           oder auf den rechten Rand der Grafik zu setzen.
    :param vertikale_position: "oben", "mitte" oder "unten" um
           die Fixierungsposition auf den oberen Rand, in die Mitte,
           oder auf den unteren Rand der Grafik zu setzen.
    :param grafik: die ursprüngliche Grafik
    :returns: die neue Grafik mit der gegebenen Fixierungsposition
    """
    h_mapping = {
        "links": "left",
        "mitte": "middle",
        "rechts": "right"
    }
    v_mapping = {
        "oben": "top",
        "mitte": "middle",
        "unten": "bottom"
    }
    return pin(h_mapping[horizontale_position],
               v_mapping[vertikale_position],
               grafik)


def ueberlagere(vordere_grafik: Grafik, hintere_grafik: Grafik) \
        -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus der zentrierten Überlagerung der zwei gegebenen Grafiken besteht.
    Die erste gegebene Grafik liegt im Vordergrund
    und die zweite im Hintergrund,
    und ihre Fixierungspositionen liegen übereinander im Zentrum.

    :param vordere_grafik: Grafik im Vordergrund
    :param hintere_grafik: Grafik im Hintergrund
    :returns: die zusammengesetzte Grafik
    """
    return overlay(vordere_grafik, hintere_grafik)


def neben(linke_grafik: Grafik, rechte_grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus dem Nebeneinanderlegen der zwei gegebenen Grafiken besteht.
    Die zwei Grafiken sind vertikal zentriert.

    :param linke_grafik: linke Grafik (im Westen)
    :param rechte_grafik: rechte Grafik (im Osten)
    :returns: die zusammengesetzte Grafik
    """
    return beside(linke_grafik, rechte_grafik)


def ueber(obere_grafik: Grafik, untere_grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik,
    die aus dem Übereinanderlegen der zwei gegebenen Grafiken besteht.
    Die zwei Grafiken sind horizontal zentriert.

    :param obere_grafik: obere Grafik (im Norden)
    :param untere_grafik: untere Grafik (im Süden)
    :returns: die zusammengesetzte Grafik
    """
    return above(obere_grafik, untere_grafik)


def drehe(grad: float, grafik: Grafik) -> Grafik:
    """
    Erzeugt eine neue Grafik, die einer Rotation der gegebenen Grafik
    um ihre Fixierungsposition im Gegenuhrzeigersinn
    um den gegebenen Winkel entspricht.

    Es kann wegen der Approximation auf die nächstgelegenen Pixel
    zu kleinen Rundungsfehlern kommen.

    :param grad: Drehwinkel, in Grad im Gegenuhrzeigersinn
    :param grafik: zu rotierende Grafik
    :returns: die neue, rotierte Grafik
    """
    return rotate(grad, grafik)



"""
Funzioni per I/O con grafiche, come visualizzarle oppure salvarle.
"""

from __future__ import annotations

from pytamaro.io import save_gif, save_graphic, show_graphic
from pytamaro.it.graphic import Grafica


def visualizza_grafica(grafica: Grafica, debug: bool = False):
    """
    Visualizza una grafica in una nuova finestra.

    Una grafica vuota non può essere mostrata; quindi chiamare questa funzione
    con essa non produce alcun affetto.

    Quando `debug` è `True`, adorna la visualizzazione con informazioni utili
    per debugging: un bordo rosso attorno alla bounding box e una croce
    giallastra attorno al punto di fissaggio.

    :param grafica: grafica da visualizzare
    :param debug: può facoltativamente essere impostato a `True` per
           sovrapporre informazioni di debug
    """
    show_graphic(grafica, debug)


def salva_grafica(nome_file: str, grafica: Grafica, debug: bool = False):
    """
    Salva una grafica come file PNG.

    Una grafica vuota non può essere salvata; quindi chiamare questa funzione
    con essa non produce alcun affetto.

    Quando `debug` è `True`, adorna la visualizzazione con informazioni utili
    per debugging: un bordo rosso attorno alla bounding box e una croce
    giallastra attorno al punto di fissaggio.

    :param nome_file: nome del file da creare (senza estensione)
    :param grafica: grafica da visualizzare
    :param debug: può facoltativamente essere impostato a `True` per
           sovrapporre informazioni di debug
    """
    save_graphic(nome_file, grafica, debug)


def salva_gif(
    nome_file: str, grafiche: list[Grafica], durata: int = 40, loop: bool = True
):
    """
    Salva una sequenza di grafiche come una GIF animata.

    Le grafiche vengono riprodotte sequenzialmente (normalmente a 25 frame al
    secondo) a ciclo continuo.

    :param nome_file: nome del file da creare (senza estensione)
    :param grafiche: lista di grafiche da salvare come GIF
    :param durata: durata in millisecondi di ciascun frame (default a 40
           millisecondi, ovvero 25 frame al secondo)
    :param loop: determina se la GIF debba riprodursi in loop indefinitamente (default a True)
    """
    save_gif(nome_file, grafiche, durata, loop)



"""
(Re-)Export Italian names for:
- basic types (Graphic and Color)
- color names
- operations
- primitves
- I/O
from this top-level module.
"""


import sys

from pytamaro.it.color import *
from pytamaro.it.color_names import *
from pytamaro.it.graphic import *
from pytamaro.it.io import *
from pytamaro.it.operations import *
from pytamaro.it.primitives import *

setattr(sys.modules["pytamaro"], "LANGUAGE", "it")



"""
Small functions used in various places to deal with graphics.
"""


import io

from PIL import Image as PILImageMod
from PIL.Image import Image as PILImage

from pytamaro.graphic import Graphic
from pytamaro.localization import translate


def ensure_size(value: float):
    """
    Raises an exception when the provided value is not valid for a
    size in pixel, being negative or equal to zero.

    :param value: the value (in pixel) to be checked
    """
    if value <= 0:
        raise ValueError(translate("INVALID_SIZE"))


def graphic_to_image(graphic: Graphic) -> PILImage:
    """
    Renders a graphic and converts it into a Pillow image.

    :param graphic: graphic to be rendered and converted
    :returns: an equivalent Pillow image
    """
    with io.BytesIO(graphic.as_image().encodeToData()) as stream:
        pil_image = PILImageMod.open(stream)
        pil_image.load()  # Ensure to make a copy of buffer
        return pil_image



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytamaro', 'pytamaro.de', 'pytamaro.it']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0.0,<10.0.0', 'skia-python>=87.4,<88.0']

setup_kwargs = {
    'name': 'pytamaro',
    'version': '0.2.0',
    'description': 'Educational library for teaching problem decompositon using graphics',
    'long_description': None,
    'author': 'Luca Chiodini',
    'author_email': 'luca@chiodini.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Tipo `Colore` e funzioni per produrre colori.
"""

from __future__ import annotations

from pytamaro.color import Color, rgb_color, rgba_color

Colore = Color
"""
Rappresenta un colore nello spazio colore RGBA (usando interi tra 0 e 255).
"""


def colore_rgb(rosso: int, verde: int, blu: int) -> Colore:
    """
    Ritorna un colore completamente opaco con le componenti indicate per il
    rosso, verde e blu.

    :param rosso: componente rosso (0 -- 255)
    :param verde: componente verde (0 -- 255)
    :param blu: componente blu (0 -- 255)
    :returns: un colore con le componenti RGB indicate
    """
    return rgb_color(rosso, verde, blu)


def colore_rgba(rosso: int, verde: int, blu: int, alpha: int) -> Colore:
    """
    Ritorna un colore con le componenti indicate per rosso, verde e blu e un
    certo grado di trasparenza determinato da `alpha`.

    :param rosso: componente rosso (0 -- 255)
    :param verde: componente verde (0 -- 255)
    :param blu: componente blu (0 -- 255)
    :param alpha: componente alpha (trasparenza), dove 0 è completamente
                  trasparente e 255 completamente opaco
    :returns: un colore con le componenti RGBA indicate
    """
    return rgba_color(rosso, verde, blu, alpha)



"""
Funktionen zur Ausgabe (Anzeigen oder Speichern) von Grafiken.
"""

from __future__ import annotations

from pytamaro.io import save_gif, save_graphic, show_graphic
from pytamaro.de.graphic import Grafik


def zeige_grafik(grafik: Grafik, debug: bool = False):
    """
    Zeige die gegebene Grafik in einem neuen Fenster an.

    Eine leere Grafik kann nicht angezeigt werden;
    deshalb hat der Aufruf dieser Funktion mit einer leeren Grafik
    keinen Effekt.

    Falls `debug` `True` ist werden auf der Grafik zusätzliche Informationen
    dargestellt, welche für das Debugging nützlich sein können.
    Ein roter Rahmen markiert die Bounding Box der Grafik,
    und ein gelbliches Kreuz gibt den Fixierpunkt an.

    :param grafik: die anzuzeigende Grafik
    :param debug: kann optional auf `True` gesetzt werden, um über der Grafik
                  Debug-Informationen darzustellen
    """
    show_graphic(grafik, debug)


def speichere_grafik(datei_name: str, grafik: Grafik, debug: bool = False):
    """
    Speichere die gegebene Grafik als PNG-Datei.

    Eine leere Grafik kann nicht gespeichert werden;
    deshalb hat der Aufruf dieser Funktion mit einer leeren Grafik
    keinen Effekt.

    Falls `debug` `True` ist werden auf der Grafik zusätzliche Informationen
    dargestellt, welche für das Debugging nützlich sein können.
    Ein roter Rahmen markiert die Bounding Box der Grafik,
    und ein gelbliches Kreuz gibt den Fixierpunkt an.

    :param datei_name: Name der zu kreierenden Datei (ohne Erweiterung)
    :param grafik: zu speichernde Grafik
    :param debug: kann optional auf `True` gesetzt werden, um über der Grafik
                  Debug-Informationen darzustellen
    """
    save_graphic(datei_name, grafik, debug)


def speichere_gif(
    datei_name: str, grafiken: list[Grafik], dauer: int = 40, loop: bool = True
):
    """
    Speichere die gegebene Sequenz von Grafiken als animierte GIF-Datei.

    Beim Anzeigen des GIFs werden die Grafiken in einer unendlichen Schleife
    animiert (normalerweise mit 25 Grafiken pro Sekunde).

    :param datei_name: Name der zu kreierenden Datei (ohne Erweiterung)
    :param grafiken: Liste der zu speichernden Grafiken
    :param dauer: Dauer jeder Grafik, in Millisekunden (Default: 40
           millisekunden, enspricht 25 Grafiken pro Sekunde)
    :param loop: bestimmt ob das GIF in einer unendlichen Schleife abgespielt
           werden soll (Default: true)
    """
    save_gif(datei_name, grafiken, dauer, loop)
